=begin
  
  sajax-ruby unofficial http://gimite.ddo.jp/pukiwiki/?sajax-ruby-en
  
  This source code is Public Domain.
  
  Based on Sajax 1.10 Ruby backend by anonym coward.
  Modified and extended by Gimite Ichikawa.
    (Some modification is suggested by senikk.)
  
=end

require "cgi"
require "date"
require "stringio"
require "erb"
require "nkf"
require "uri"


class Sajax

  def initialize(params= {})
    
    @debug_mode= params[:debug_mode] || false
    @remote_uri= params[:remote_uri] || ENV["REQUEST_URI"]
    @kcode= params[:kcode] || $KCODE
    @mb_support= params.has_key?(:mb_support) ? params[:mb_support] : (RUBY_VERSION >= "1.8.2")
    @request_type= (params[:request_type] || "GET").to_s().upcase()
    
    if !(@request_type=~/^(GET|POST)$/)
      raise(ArgumentError, "Unsupported request type: "+@request_type)
    end
    
    @export_list= {}
    @js_has_been_shown= false
    @cgi= CGI.new()
    
  end

  def handle_client_request()
    
    func_name= decode_request(@cgi.params['rs'].to_s())
    return if func_name == ""
    
    out= @cgi.header({
      "Expires" => "Mon, 26 Jul 1997 05:00:00 GMT",
      "Last-Modified" => DateTime.now.new_offset.strftime("%a, %d %m %H:%M:%S GMT"), # Always modified.
      "Cache-Control" => "no-cache, must-revalidate", # For HTTP/1.1.
      "Pragma" => "no-cache", # For HTTP/1.0.
    })

    if !@export_list.has_key?(func_name)
      out << ("-:%s not callable" % encode_result(func_name))
    else
      rsargs = @cgi.params['rsargs[]'].map(){ |s| decode_request(s) }
      result = @export_list[func_name].call(*rsargs)
      out << ("+:%s" % encode_result(result))
    end
    
    print_and_exit(out)
    
  end
  
  def print_and_exit(out)
    print(out)
    exit()
  end
  
  def decode_request(str)
    if @mb_support
      return unicode_to_kcode(unescape_unicode(str))
    else
      return str
    end
  end
  
  def encode_result(str)
    if @mb_support
      return escape_unicode(kcode_to_unicode(str))
    else
      return URI.escape(str)
    end
  end
  
  def unescape_unicode(str)
    return str.gsub(/%u([0-9a-fA-F]{1,4})|./n) do
      $1 ? [$1].pack("H*") : "\0"+$&
    end
  end
  
  def escape_unicode(str)
    return str.gsub(/(.)(.)/n) do
      if $1=="\0" && $2!="%"
        $2
      else
        "%u" + $&.unpack("H*")[0].upcase()
      end
    end
  end
  
  def unicode_to_kcode(str)
    c= self.nkf_kcode
    return NKF.nkf("-#{c}W16B", str)
  end
  
  def kcode_to_unicode(str)
    c= self.nkf_kcode.upcase
    return NKF.nkf("-w16b#{c}", str.to_s())
  end
  
  def nkf_kcode
    case @kcode
      when /^e/i; "e"   # EUC-JP
      when /^s/i; "s"   # Shift_JIS
      else      ; "w80" # UTF-8N (UTF-8 without BOM)
    end
  end
  
  def common_js
    
    js_debug_mode= @debug_mode ? "true" : "false"
    js_request_type= @request_type
    js_remote_uri= @remote_uri
    
    return <<-"EOS"
      
      // remote scripting library
      // (c) copyright 2005 modernmethod, inc
      
      var sajax_debug_mode= #{js_debug_mode};
      var sajax_request_type= "#{js_request_type}";

      function sajax_debug(text) {
        if (sajax_debug_mode)
          alert("RSD: " + text)
      }
      
      function sajax_init_object() {
        sajax_debug("sajax_init_object() called..")
        var A;
        try {
          A = new ActiveXObject("Msxml2.XMLHTTP");
        } catch (e) {
          try {
            A = new ActiveXObject("Microsoft.XMLHTTP");
          } catch (oc) {
            A = null;
          }
        }
        if(!A && typeof XMLHttpRequest != "undefined")
          A = new XMLHttpRequest();
        if (!A)
          sajax_debug("Could not create connection object.");
        return A;
      }
      
      function sajax_do_call(func_name, args) {
        var i, x, n;
        var uri;
        var post_data;
        uri = "#{js_remote_uri}";
        if (sajax_request_type == "GET") {
          if (uri.indexOf("?") == -1) 
            uri = uri + "?rs=" + escape(func_name);
          else
            uri = uri + "&rs=" + escape(func_name);
          for (i = 0; i < args.length-1; i++) 
            uri = uri + "&rsargs[]=" + escape(args[i]);
          uri = uri + "&rsrnd=" + new Date().getTime();
          post_data = null;
        } else {
          post_data = "rs=" + escape(func_name);
          for (i = 0; i < args.length-1; i++) 
            post_data = post_data + "&rsargs[]=" + escape(args[i]);
        }
        x = sajax_init_object();
        x.open(sajax_request_type, uri, true);
        if (sajax_request_type == "POST") {
          x.setRequestHeader("Method", "POST " + uri + " HTTP/1.1");
          x.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");
        }
        x.onreadystatechange = function() {
          if (x.readyState != 4) 
            return;
          sajax_debug("received " + x.responseText);
          var status = x.responseText.charAt(0);
          var data = unescape(x.responseText.substring(2));
          if (status == "-") 
            alert("Error: " + data);
          else  
            args[args.length-1](data);
        }
        x.send(post_data);
        sajax_debug(func_name + " uri = " + uri + "/post = " + post_data);
        sajax_debug(func_name + " waiting..");
        delete x;
      }
    
    EOS
    
  end
  
  def one_stub(func_name)
    return <<-"EOS"
      // Wrapper for #{func_name}.
      
      function x_#{func_name}() {
        sajax_do_call("#{func_name}",
          x_#{func_name}.arguments);
      }
      
    EOS
  end
  
  def javascript
    html= ""
    if !@js_has_been_shown
      html << common_js
      @js_has_been_shown= true
    end
    for func_name in @export_list.keys
      html << one_stub(func_name)
    end
    return html
  end
  
  def export(name, proc= nil, &block)
    @export_list[name] = proc || block
  end

  attr_reader(:debug_mode, :remote_uri, :kcode, :mb_support, :cgi)

end
